@@ -11,12 +11,14 @@ and manipulating text within those ranges. The main components are:
 
 import re
 from collections.abc import Sequence
-from typing import NamedTuple
+from typing import NamedTuple, TypeAlias
 from functools import total_ordering
+from dataclasses import dataclass, field
 
 
 from cedarscript_ast_parser import Marker, RelativeMarker, RelativePositionType, MarkerType, BodyOrWhole
-from .indentation_kit import get_line_indent_count
+
+from .line_kit import get_line_indent_count_from_lines
 
 MATCH_TYPES = ('exact', 'stripped', 'normalized', 'partial')
 
@@ -40,24 +42,49 @@ class RangeSpec(NamedTuple):
     indent: int = 0
 
     def __str__(self):
-        """Return a string representation of the RangeSpec."""
         return (f'{self.start}:{self.end}' if self.as_index is None else f'%{self.as_index}') + f'@{self.indent}'
 
     def __lt__(self, other):
-        """Compare if this range is strictly before another range."""
-        return self.end < other.start
+        """Compare if this range is strictly before another range or int."""
+        match other:
+            case int():
+                return self.end <= other
+            case RangeSpec():
+                return self.end <= other.start
 
     def __le__(self, other):
         """Compare if this range is before or adjacent to another range."""
-        return self.end <= other.start
+        match other:
+            case int():
+                return self.end <= other - 1
+            case RangeSpec():
+                return self.end <= other.start - 1
 
     def __gt__(self, other):
         """Compare if this range is strictly after another range."""
-        return self.start > other.end
+        match other:
+            case int():
+                return self.start > other
+            case RangeSpec():
+                return self.start >= other.end
 
     def __ge__(self, other):
         """Compare if this range is after or adjacent to another range."""
-        return self.start >= other.end
+        match other:
+            case int():
+                return self.start >= other
+            case RangeSpec():
+                return self.start >= other.end - 1
+
+    def __contains__(self, item):
+        match item:
+            case int():
+                return self.start <= item < self.end
+            case RangeSpec():
+                return (
+                    self == RangeSpec.EMPTY or
+                    item != RangeSpec.EMPTY and self.start <= item.start and item.end <= self.end
+                )
 
     @property
     def line_count(self):
@@ -92,7 +119,7 @@ class RangeSpec(NamedTuple):
 
     def write(self, src: Sequence[str], target: Sequence[str]):
         """Write the source lines into the target sequence at the index position specified by this range."""
-        target[self.start:self.end] = src
+        target[self.start:self.end if self.end >= 0 else len(target)] = src
 
     def delete(self, src: Sequence[str]) -> Sequence[str]:
         """Delete the lines specified by this range from the source sequence and return the deleted lines."""
@@ -143,15 +170,12 @@ class RangeSpec(NamedTuple):
             - The method prioritizes match types in the order: exact, stripped, normalized, partial.
             - The offset is considered separately for each match type.
         """
-        search_start_index, search_end_index, _ = search_range if search_range is not None else (0, -1, 0)
+        search_start_index, search_end_index, _ = search_range if search_range is not None else RangeSpec.EMPTY
         search_line = search_term.value
-        assert search_line, "Empty marker"
         assert search_term.type == MarkerType.LINE, f"Invalid marker type: {search_term.type}"
 
-        matches = {t: [] for t in MATCH_TYPES}
 
-        stripped_search = search_line.strip()
-        normalized_search_line = cls.normalize_line(stripped_search)
+        matches = {t: [] for t in MATCH_TYPES}
 
         if search_start_index < 0:
             search_start_index = 0
@@ -167,44 +191,105 @@ class RangeSpec(NamedTuple):
             f"must be less than or equal to line count ({len(lines)})"
         )
 
+        marker_subtype = (search_term.marker_subtype or "string").casefold()
+        assert search_line is not None or marker_subtype == 'empty', "Empty marker"
+
+        # Handle special marker subtypes that don't use normal line matching
+        match marker_subtype:
+            case 'number':  # Match by line number relative to search range
+                relative_index = search_line - 1
+                if search_range:
+                    # Make index relative to search range start
+                    index = search_range.start + relative_index
+                    if not (0 <= index <= len(lines)):
+                        raise ValueError(
+                            f"Line number {search_line} out of bounds "
+                            f"(must be in interval [1, {len(lines) + 1}] relative to context)"
+                        )
+                else:
+                    index = relative_index
+                    if not (0 <= index < len(lines)):
+                        raise ValueError(
+                            f"Line number {search_line} out of bounds "
+                            f"(must be in interval [1, {len(lines)}])"
+                        )
+                reference_indent = get_line_indent_count_from_lines(lines, index)
+                index += calc_index_delta_for_relative_position(search_term)
+                return cls(index, index, reference_indent)
+
+            case 'regex':  # Match using regex pattern
+                try:
+                    pattern = re.compile(search_line)
+                except re.error as e:
+                    raise ValueError(f"Invalid regex pattern '{search_line}': {e}")
+
+            case _:  # Default string matching modes
+                pattern = None
+                stripped_search = search_line.strip() if search_line else ""
+                normalized_search_line = cls.normalize_line(stripped_search)
+
+        # Find all matching lines based on marker subtype
         for i in range(search_start_index, search_end_index):
+            reference_indent = get_line_indent_count_from_lines(lines, i)
             line = lines[i]
-            reference_indent = get_line_indent_count(line)
+            stripped_line = line.strip()
+            normalized_line = cls.normalize_line(line)
 
-            # Check for exact match
-            if search_line == line:
-                matches['exact'].append((i, reference_indent))
+            match marker_subtype:
+                case 'empty':
+                    if not line or not stripped_line:
+                        matches['stripped'].append((i, reference_indent))
 
-            # Check for stripped match
-            elif stripped_search == line.strip():
-                matches['stripped'].append((i, reference_indent))
+                case 'indent-level':
+                    if reference_indent == search_line: # TODO Calc indent level
+                        matches['exact'].append((i, reference_indent))
 
-            # Check for normalized match
-            elif normalized_search_line == cls.normalize_line(line):
-                matches['normalized'].append((i, reference_indent))
+                case 'regex':
+                    if pattern.search(line) or pattern.search(stripped_line):
+                        matches['exact'].append((i, reference_indent))
 
-            # Last resort!
-            elif normalized_search_line.casefold() in cls.normalize_line(line).casefold():
-                matches['partial'].append((i, reference_indent))
+                case 'prefix':
+                    if stripped_line.startswith(stripped_search):
+                        matches['exact'].append((i, reference_indent))
+                    elif normalized_line.startswith(normalized_search_line):
+                        matches['normalized'].append((i, reference_indent))
+
+                case 'suffix':
+                    if stripped_line.endswith(stripped_search):
+                        matches['exact'].append((i, reference_indent))
+                    elif normalized_line.endswith(normalized_search_line):
+                        matches['normalized'].append((i, reference_indent))
+
+                case 'string' | _:  # Default string matching
+                    if search_line == line:
+                        matches['exact'].append((i, reference_indent))
+                    elif stripped_search == stripped_line:
+                        matches['stripped'].append((i, reference_indent))
+                    elif normalized_search_line == normalized_line:
+                        matches['normalized'].append((i, reference_indent))
+                    elif normalized_search_line.casefold() in normalized_line.casefold():
+                        matches['partial'].append((i, reference_indent))
 
         offset = search_term.offset or 0
+        max_match_count = max([len(m) for m in matches.values()])
         for match_type in MATCH_TYPES:
             match_type_count = len(matches[match_type])
             if search_term.offset is None and match_type_count > 1:
                 raise ValueError(
-                    f"There are {match_type_count} lines matching `{search_term.value}`. "
-                    "Suggestions: 1) Try using a *different line* as marker (a couple lines before or after the one "
-                    "you tried); 2) If you wanted to *REPLACE* line, "
-                    "try instead to replace a *SEGMENT* of a couple of lines."
+                    f"Line marker `{search_term.value}` is ambiguous (found {match_type_count} lines matching it) "
+                    "Suggestions: 1) To disambiguate, try using a *different line* as marker (a couple lines before or "
+                    "after the one you tried); 2) If you wanted to *REPLACE* line, try instead to replace a *SEGMENT* "
+                    "a couple of lines long."
                     # f"Add an `OFFSET` (after the line marker) and a number between 0 and {match_type_count - 1}
                     # to determine how many to skip. "
                     # f"Example to reference the *last* one of those:
                     # `LINE '{search_term.value.strip()}' OFFSET {match_type_count - 1}`"
                     # ' (See `offset_clause` in `<grammar.js>` for details on OFFSET)'
                 )
-            if match_type_count and (search_term.offset or 0) >= match_type_count:
+
+            if match_type_count and offset >= max_match_count:
                 raise ValueError(
-                    f"There are only {match_type_count} lines matching `{search_term.value}`, "
+                    f"There are only {max_match_count} lines matching `{search_term.value}`, "
                     f"but 'OFFSET' was set to {search_term.offset} (you can skip at most {match_type_count-1} of those)"
                 )
 
@@ -216,16 +301,7 @@ class RangeSpec(NamedTuple):
                     case 'partial':
                         print(f"Note: Won't accept {match_type} match at index {index} for {search_term}")
                         continue
-                if isinstance(search_term, RelativeMarker):
-                    match search_term.qualifier:
-                        case RelativePositionType.BEFORE:
-                            index += -1
-                        case RelativePositionType.AFTER:
-                            index += 1
-                        case RelativePositionType.AT:
-                            pass
-                        case _ as invalid:
-                            raise ValueError(f"Not implemented: {invalid}")
+                index += calc_index_delta_for_relative_position(search_term)
                 return cls(index, index, reference_indent)
 
         return None
@@ -234,7 +310,30 @@ class RangeSpec(NamedTuple):
 RangeSpec.EMPTY = RangeSpec(0, -1, 0)
 
 
-class IdentifierBoundaries(NamedTuple):
+def calc_index_delta_for_relative_position(marker: Marker):
+    match marker:
+        case RelativeMarker(qualifier=RelativePositionType.BEFORE):
+            return -1
+        case RelativeMarker(qualifier=RelativePositionType.AFTER):
+            return 1
+        case RelativeMarker(qualifier=RelativePositionType.AT):
+            pass
+        case RelativeMarker(qualifier=invalid):
+            raise ValueError(f"Not implemented: {invalid}")
+        case _:
+            pass
+    return 0
+
+
+class ParentInfo(NamedTuple):
+    parent_name: str
+    parent_type: str
+
+
+ParentRestriction: TypeAlias = RangeSpec | str | None
+
+@dataclass
+class IdentifierBoundaries:
     """
     Represents the boundaries of an identifier in code, including its whole range and body range.
 
@@ -249,7 +348,12 @@ class IdentifierBoundaries(NamedTuple):
     whole: RangeSpec
     body: RangeSpec | None = None
     docstring: RangeSpec | None = None
-    decorators: list[RangeSpec] = []
+    decorators: list[RangeSpec] = field(default_factory=list)
+    parents: list[ParentInfo] = field(default_factory=list)
+
+    def append_decorator(self, decorator: RangeSpec):
+        self.decorators.append(decorator)
+        self.whole = self.whole._replace(start = min(self.whole.start, decorator.start))
 
     def __str__(self):
         return f'IdentifierBoundaries({self.whole} (BODY: {self.body}) )'
@@ -269,6 +373,32 @@ class IdentifierBoundaries(NamedTuple):
         """Return the 1-indexed end line of the whole identifier."""
         return self.whole.end
 
+    def match_parent(self, parent_restriction: ParentRestriction) -> bool:
+        match parent_restriction:
+            case None:
+                return True
+            case RangeSpec():
+                return self.whole in parent_restriction
+            case str() as parent_spec:
+                # Parent chain matching: Handle dot notation for parent relationships
+                parent_chain = parent_spec.split('.')
+                if len(parent_chain) == 1:
+                    # Simple case - just check if name is any of the parents
+                    return parent_spec in [p.parent_name for p in self.parents]
+                parent_chain = [p for p in parent_chain if p]
+                if len(parent_chain) > len(self.parents):
+                    return False
+                # len(parent_chain) <= len(self.parents)
+                # Check parent chain partially matches (
+                #   sub-chain match when there are fewer items in 'parent_chain' than in 'self.parents'
+                # )
+                return all(
+                    expected == actual.parent_name
+                    for expected, actual in zip(parent_chain, self.parents)
+                )
+            case _:
+                raise ValueError(f'Invalid parent restriction: {parent_restriction}')
+
     def location_to_search_range(self, location: BodyOrWhole | RelativePositionType) -> RangeSpec:
         """
         Convert a location specifier to a RangeSpec for searching.
@@ -294,9 +424,9 @@ class IdentifierBoundaries(NamedTuple):
                 return RangeSpec(self.whole.start, self.whole.start, self.whole.indent)
             case RelativePositionType.AFTER:
                 return RangeSpec(self.whole.end, self.whole.end, self.whole.indent)
-            case RelativePositionType.INSIDE_TOP:
+            case RelativePositionType.INTO_TOP:
                 return RangeSpec(self.body.start, self.body.start, self.body.indent)
-            case RelativePositionType.INSIDE_BOTTOM:
+            case RelativePositionType.INTO_BOTTOM:
                 return RangeSpec(self.body.end, self.body.end, self.body.indent)
             case _ as invalid:
                 raise ValueError(f"Invalid: {invalid}")
