@@ -8,7 +8,7 @@ manipulations, which are useful for tasks such as code analysis and transformati
 
 from collections.abc import Sequence
 from typing import Protocol, runtime_checkable
-from os import PathLike
+from os import PathLike, path
 
 from cedarscript_ast_parser import Marker, RelativeMarker, RelativePositionType, Segment, MarkerType, BodyOrWhole
 from .range_spec import IdentifierBoundaries, RangeSpec
@@ -24,7 +24,7 @@ def read_file(file_path: str | PathLike) -> str:
     Returns:
         str: The contents of the file as a string.
     """
-    with open(file_path, 'r') as file:
+    with open(path.normpath(file_path), 'r') as file:
         return file.read()
 
 
@@ -36,7 +36,7 @@ def write_file(file_path: str | PathLike, lines: Sequence[str]):
         file_path (str | PathLike): The path to the file to be written.
         lines (Sequence[str]): The lines to be written to the file.
     """
-    with open(file_path, 'w') as file:
+    with open(path.normpath(file_path), 'w') as file:
         file.writelines([line + '\n' for line in lines])
 
 
@@ -170,6 +170,9 @@ def segment_to_search_range(
     """
     assert len(lines), "`lines` is empty"
 
+    match search_range:
+        case None:
+            search_range = RangeSpec.EMPTY
     start_match_result = RangeSpec.from_line_marker(lines, start_relpos, search_range)
     assert start_match_result, (
         f"Unable to find segment start: {start_relpos}; Try: "
@@ -178,12 +181,17 @@ def segment_to_search_range(
     )
 
     start_index_for_end_marker = start_match_result.as_index
-    match start_relpos:
-        case RelativeMarker(qualifier=RelativePositionType.AFTER):
-            start_index_for_end_marker += -1
-    end_match_result = RangeSpec.from_line_marker(lines, end_relpos, RangeSpec(
-        start_index_for_end_marker, search_range.end, start_match_result.indent
-    ))
+    search_range_for_end_marker = search_range
+    if end_relpos.marker_subtype != 'number':
+        match start_relpos:
+            case RelativeMarker(qualifier=RelativePositionType.AFTER):
+                start_index_for_end_marker += -1
+                search_range_for_end_marker = RangeSpec(
+                    start_index_for_end_marker,
+                    search_range.end,
+                    start_match_result.indent
+                )
+    end_match_result = RangeSpec.from_line_marker(lines, end_relpos, search_range_for_end_marker)
     assert end_match_result, (
         f"Unable to find segment end: {end_relpos}; Try: "
         f"1) Using *exactly* the same characters from source; or "
